(*
Author:  Christina Kohl <christina.kohl@uibk.ac.at> (2023)
License: LGPL (see file COPYING.LESSER)
*)

text\<open>Check functions for confluence and commutation via development closed critical pairs.
Analogous to Parallel_Closed_Impl.thy\<close>

theory Development_Closed_Impl
imports
  Development_Closed
  Critical_Pairs_Impl
  Check_Joins
  Commutation
begin

fun is_development_closed_cp where
  "is_development_closed_cp checkStep listSteps R _ (False, s, t) = checkStep s t"
| "is_development_closed_cp checkStep listSteps R (Some n) (True, s, t) =
    (\<not> Option.is_none (List.find (\<lambda>v. checkStep s v) (reachable_terms R t n)))"
| "is_development_closed_cp checkStep listSteps R None (True, s, t) =
    (\<not> Option.is_none (List.find (\<lambda>v. checkStep s v) (listSteps t)))"


definition check_development_closed where
  "check_development_closed ren R n = do {
     check_wf_trs R;
     check_left_linear_trs R;
     let is_step = is_mstep R; 
     let steps_list = mstep_rewrite R;
     check_allm (\<lambda> (b, s, t). do {
       check (is_development_closed_cp is_step steps_list R n (b, s, t))
         (showsl_lit (STR ''the critical pair '') \<circ> showsl s \<circ> showsl_lit (STR '' <- . -> '') \<circ> showsl t \<circ>
          showsl_lit (STR '' is not (almost) development closed within '') \<circ> showsl n \<circ> showsl_lit (STR '' steps.''))
     }) (critical_pairs_impl ren R R)
     }  <+? (\<lambda>s. s \<circ> showsl_lit (STR ''\<newline>hence the following TRS is not development closed\<newline>'') \<circ> showsl_trs R)"


lemma check_development_closed:
  assumes "isOK(check_development_closed ren R n)"
  shows "CR (rstep (set R))"
proof (rule mstep_closed_imp_CR)
  from assms show "left_linear_trs (set R)" 
    unfolding check_development_closed_def by auto 
  from assms have "wf_trs (set R)" 
    unfolding check_development_closed_def by auto
  thus "(l, r) \<in> set R \<Longrightarrow> is_Fun l \<and> vars_term r \<subseteq> vars_term l" for l r
    unfolding wf_trs_def by auto
  show "\<And>s t. (False, s, t) \<in> critical_pairs ren (set R) (set R) \<Longrightarrow> (s, t) \<in> mstep (set R)"
  proof -
    fix s t
    assume "(False, s, t) \<in> critical_pairs ren (set R) (set R)"
    with assms have dc:"is_development_closed_cp (is_mstep R) (mstep_rewrite R) R n (False, s, t)" 
      unfolding check_development_closed_def by auto
    with dc show "(s, t) \<in> mstep (set R)" 
      using varcond is_development_closed_cp.simps[of "is_mstep R"] by auto
  qed
  show "\<And>s t. (True, s, t) \<in> critical_pairs ren (set R) (set R) \<Longrightarrow> \<exists>v. (s, v) \<in> mstep (set R) \<and> (t, v) \<in> (rstep (set R))\<^sup>*"
  proof -
    fix s t
    assume "(True, s, t) \<in> critical_pairs ren (set R) (set R)"
    with assms have dc:"is_development_closed_cp (is_mstep R) (mstep_rewrite R) R n (True, s, t)" 
      unfolding check_development_closed_def by auto
    show "\<exists>v. (s, v) \<in> mstep (set R) \<and> (t, v) \<in> (rstep (set R))\<^sup>*" proof(cases n)
      case None
      with dc have "\<not> Option.is_none (List.find (\<lambda>v. is_mstep R s v) (mstep_rewrite R t))" 
        using varcond is_development_closed_cp.simps by auto
      then obtain v where "List.find (\<lambda>v. is_mstep R s v) (mstep_rewrite R t) = Some v"
        by force
      then have "(s,v) \<in> mstep (set R)" and "v \<in> set (mstep_rewrite R t)"
        unfolding find_Some_iff by auto
      then show ?thesis
        using mstep_imp_rsteps mstep_rewrite_mstep by blast 
    next
      case (Some i)
      with dc have "\<not> Option.is_none (List.find (\<lambda>v. is_mstep R s v) (reachable_terms R t i))" 
        using varcond is_development_closed_cp.simps by auto
      then obtain v where "List.find (\<lambda>v. is_mstep R s v) (reachable_terms R t i) = Some v"
        by force
      then have "(s,v) \<in> mstep (set R)" and "v \<in> set (reachable_terms R t i)"
        unfolding find_Some_iff by auto
      then show ?thesis
        using reachable_terms by blast
    qed
  qed
qed


definition check_development_closed_comm where
  "check_development_closed_comm ren R S n = do {
     check_wf_trs R;
     check_wf_trs S;
     check_left_linear_trs R;
     check_left_linear_trs S;
     check_allm (\<lambda> (b, s, t). do {
       check (is_mstep_join S R n s t)
         (showsl_lit (STR ''the critical pair '') \<circ> showsl s \<circ> showsl_lit (STR '' R<- . ->S '') \<circ> showsl t \<circ>
          showsl_lit (STR '' is not almost development closed within '') \<circ> showsl n \<circ> showsl_lit (STR '' steps.''))
     }) (critical_pairs_impl ren S R);
     check_allm (\<lambda> (b, s, t). do {
       check (b \<or> is_mstep R s t)
         (showsl_lit (STR ''the inner critical pair '') \<circ> showsl s \<circ> showsl_lit (STR '' S<- . ->R '') \<circ> showsl t \<circ>
          showsl_lit (STR '' is not closed with multi-step over R'') \<circ> showsl n)
     }) (critical_pairs_impl ren R S)
     }  <+? (\<lambda>s. s \<circ> showsl_lit (STR ''\<newline>hence the almost development closed check for the following TRSs could not be proven\<newline>R: '') \<circ> 
        showsl_trs R o showsl_lit (STR ''\<newline>\<newline>S: '') o showsl_trs S)"

lemma check_development_closed_comm:
  assumes "isOK(check_development_closed_comm ren R S n)"
  shows "sig_commute F (set R) (set S)"
proof (intro commute_imp_sig_commute mstep_closed_imp_commute)
  note assms = assms[unfolded check_development_closed_comm_def, simplified, unfolded wf_trs_def']
  from assms show "\<And>l r. (l, r) \<in> set R \<Longrightarrow> is_Fun l \<and> vars_term r \<subseteq> vars_term l"
    by auto 
  from assms show "\<And>l r. (l, r) \<in> set S \<Longrightarrow> is_Fun l \<and> vars_term r \<subseteq> vars_term l"
    by auto 
  from assms show "left_linear_trs (set R)" "left_linear_trs (set S)" by auto
  from assms show "(b, p, q) \<in> critical_pairs ren (set S) (set R) \<Longrightarrow> \<exists>v. (p, v) \<in> mstep (set S) \<and> (q, v) \<in> (rstep (set R))\<^sup>*" for b p q
    using is_mstep_join by (metis (mono_tags, lifting) case_prodD is_mstep) 
  from assms show "\<And>p q. (False, p, q) \<in> critical_pairs ren (set R) (set S) \<Longrightarrow> (p,q) \<in> mstep (set R)"
    using is_mstep by (smt (verit) Ball_set_list_all case_prodD) 
qed

end
